Option Explicit

' Released: 2010 March 24

' OpenFileGrumbler

' This program builds a list of open files, and then (1) emails a
' report to the administrator and (2) emails a message to each user
' with open files.  Intended to be used as a nightly auto-run script,
' to remind people to close their files before they go home.

' Provided free-of-charge, without restriction on (re)distribution,
' modification, or use.
'
' This program is provided "AS IS" and WITHOUT WARRANTY of any kind.
' USE STRICTLY AT YOUR OWN RISK.

' Written by Ben Scott

' ======================================================================
' globals

' scripting fundamentals
Dim WshShell
Dim WshNetwork
Dim FSO
Const WshShellRunHidden = 0
Const WshShellRunWait = True
Const ForReading = 1, ForWriting = 2 ' for FSO.OpenTextFile

' special characters
Const dq = """" ' double quote (")
Const emptyString = ""

' configuration (read from file)
Dim fileServer		' name of file server to check for open files
Dim mailServer		' name of SMTP mail relay server
Dim userDomain		' appended to user names to generate email addresses
Dim adminEmail		' summary report will be sent to this address
Dim fromEmail		' used for the "From" address of every message
Dim userSubj		' subject line for email sent to users
Dim bodyFileName	' name of file containing user message body text
Dim userSkipListFileName' name of file containing list of users to exclude
Dim fileSkipListFileName' name of file containing list of files to exclude

' counters
Dim fileCount ' open files (files open more than once count each open)
Dim userCount ' number of users with open files
Dim queryFailures ' open things which can't be queried for their user and/or file name
Dim missingFiles ' open things which don't exist in filesystem (mainly pipes)

' lists of things to skip
Dim fileSkipList
Dim userSkipList

' collections and their iterators
Dim offenders  ' key = username; value = openFiles Dictionary object
Dim offender
Dim openFiles  ' per user; key = file path; value = count of opens
Dim openFile

' misc
Dim Debugging	' if TRUE, debugging mode is enabled
Dim blatBase	' basic BLAT command, same for all mail
Dim bodyText	' body text of message sent to users
Dim hostname	' name of computer script is running on

' ======================================================================
' subs and functions

' ----------------------------------------------------------------------
Sub Die (msg) ' abort with message
WScript.Echo msg
WScript.Quit 1 ' non-zero exit status indicates failure
End Sub ' Die

' ----------------------------------------------------------------------
Sub DMsg (msg) ' debug message
If Debugging Then WScript.Stderr.WriteLine msg
End Sub ' DMsg

' ----------------------------------------------------------------------
Sub RunWait (cmd) ' run a command, waiting until it finishes
WshShell.Run cmd, WshShellRunHidden, WshShellRunWait
End Sub

' ----------------------------------------------------------------------
Sub ReadMasterConfigFile (fileName)

Dim confFile

DMsg "reading config file..."

Set confFile = FSO.OpenTextFile(fileName, ForReading, False)

fileServer = confFile.ReadLine
mailServer = confFile.ReadLine
userDomain = confFile.ReadLine
adminEmail = confFile.ReadLine
fromEmail = confFile.ReadLine
userSubj = confFile.ReadLine
bodyFileName = confFile.ReadLine
userSkipListFileName = confFile.ReadLine
fileSkipListFileName = confFile.ReadLine

confFile.Close

End Sub ' ReadMasterConfigFile 

' ----------------------------------------------------------------------
Function DictListFromFile (fileName)
' Reads a text file into a dictionary.  Each line in the file is
' used as the dictionary key; the dictionary value is the count of lines
' matching that key exactly.  Used mainly to implement a poor man's
' unordered dynamic list.

Dim listFile
Dim dictList
Dim line

Set dictList = CreateObject("Scripting.Dictionary")

Set listFile = FSO.OpenTextFile(fileName, ForReading, False)

Do While Not listfile.AtEndOfStream
	line = listFile.ReadLine
	If Not dictList.Exists(line) Then
		dictList.Add line, 1
	Else
		dictList.Item(line) = dictList.Item(line) + 1
	End If
Loop

listFile.Close

Set DictListFromFile = dictList

End Function

' ----------------------------------------------------------------------
Sub BlatMail (subj, addr)

RunWait blatBase & " -s " & dq & subj & dq & " -to " & addr

End Sub 'BlatMail

' ----------------------------------------------------------------------
Sub AddOpenFile (user, path)

Dim userFileList

If ShouldSkip(user, path) Then Exit Sub

fileCount = fileCount + 1

' find or add the open file list for this user
If Not offenders.Exists(user) Then
	userCount = userCount + 1
	Set userFileList = CreateObject("Scripting.Dictionary")
	offenders.Add user, userFileList
Else
	Set userFileList = offenders.Item(user)
End If
	
' create or increment the count for this file
If Not userFileList.Exists(path) Then
	userFileList.Add path, 1
Else
	userFileList.Item(path) = userFileList.Item(path) + 1
End If

End Sub ' addOpenFile

' ----------------------------------------------------------------------
Function ShouldSkip(user, path)

' early exit means "yes", we should skip
ShouldSkip = True

' if in skip list, skip
If userSkipList.Exists(user) Then Exit Function
If fileSkipList.Exists(path) Then Exit Function

' if a machine account, skip
' machine accounts end in dollar sign
If Right(user, 1) = "$" Then Exit Function

' skip folders
If FSO.FolderExists(path) Then Exit Function

' things which do not exist in filesystem are counted
' they are NOT skipped; this is just informational
If Not FSO.FileExists(path) Then
	missingFiles = missingFiles + 1
	DMsg "missing file open: " & path
End If

' everything passed, do not skip
ShouldSkip = False

End Function ' ShouldSkip

' ======================================================================
' main program

' ----------------------------------------------------------------------
' variables

Dim bodyFile  ' message body text file
Dim userEmail ' user email address
Dim sendTo    ' an email address to send to

' temp file
Dim tempDirObj
Dim tempDirPath
Dim tempName
Dim tempFileName
Dim tempFile

' ----------------------------------------------------------------------
' initialize fundamentals

Set WshShell = WScript.CreateObject("WScript.Shell")
Set WshNetwork = WScript.CreateObject("WScript.Network")
Set FSO = CreateObject("Scripting.FileSystemObject")
hostname = WshNetwork.ComputerName

' ----------------------------------------------------------------------
' get config file from command line and config from config file

Dim Args
Dim masterConfigFileName

Set Args = WScript.Arguments
If (Args.Count < 1) Then Die "Missing program arguments (command line)"
If (Args.Count > 2) Then Die "Too many program arguments"

masterConfigFileName = Args(0)

If (Args.Count = 2) Then
	If (Args(1) <> "debug") Then Die "Bad second argument"
	Debugging = True
Else
	Debugging = False
End If

DMsg "script started"

' first argument is path and name of master config file
ReadMasterConfigFile(masterConfigFileName)

' ----------------------------------------------------------------------
' initialize derived configuration

' zero counters
fileCount = 0
userCount = 0
queryFailures = 0
missingFiles = 0

' read skip lists
Set userSkipList = DictListFromFile(userSkipListFileName)
Set fileSkipList = DictListFromFile(fileSkipListFileName)

' create temp file, for BLAT
DMsg "initializing temp file..."
Set tempDirObj = FSO.GetSpecialFolder(2)
tempDirPath = tempDirObj.Path
tempName = FSO.GetTempName
tempFileName = FSO.BuildPath(tempDirPath, tempName)

' build BLAT base command
blatBase = "BLAT " & tempFileName _
	& " -server " & mailServer _
	& " -f " & fromEmail

' get message body text
DMsg "reading user message body..."
Set bodyFile = FSO.OpenTextFile(bodyFileName, ForReading, False)
bodyText = bodyFile.ReadAll
bodyFile.Close

' ----------------------------------------------------------------------
' build list of users and their open files

Dim user
Dim path

DMsg "getting list of open files..."
Set offenders = CreateObject("Scripting.Dictionary")
Set openFiles = GetObject("WinNT://" & fileServer & "/LanmanServer").Resources
For Each openFile in openFiles
	user = emptyString
	path = emptyString
	On Error Resume Next
		user = openFile.user
		path = openFile.path
	On Error GoTo 0
	If ((User = emptyString) OR (path = emptyString)) Then
		' once in a great while, accessing .user or .path will throw a "property cannot be found in the cache" error
		' I suspect these are files which were open when openFiles was built, but have since been closed
		' all we can do is count them, since the name itself is usually unavailable
		DMsg "unable to get user or path for an open object"
		queryFailures = queryFailures + 1
	Else
		AddOpenFile user, path
	End If
Next ' openFile

' ----------------------------------------------------------------------
' master report, to admin

DMsg "generating admin report..."

Set tempFile = FSO.OpenTextFile(tempFileName, ForWriting, True)

tempFile.WriteLine "Open file report"
tempFile.WriteLine "Generated " & Now
tempFile.WriteLine "Generated on computer " & hostname
tempFile.WriteLine "For server " & fileServer
tempFile.WriteLine "Admin contact: " & adminEmail
tempFile.WriteLine userCount & " users with " & fileCount & " files open"
If (missingFiles > 0) Then tempFile.WriteLine missingFiles & " of those files were reported not to exist (missingFiles)"
If (queryFailures > 0) Then tempFile.WriteLine queryFailures & " of those files could not be queried (queryFailures)"
tempFile.WriteLine

For Each offender in offenders.Keys
	tempFile.WriteLine "User " & offender & ":"
	For Each openFile in offenders.Item(offender).Keys
		tempFile.WriteLine "  " & openFile
	Next
	tempFile.WriteLine
Next
tempFile.WriteLine "End of open file report"

tempFile.Close

BlatMail "Open file report", adminEmail

' ----------------------------------------------------------------------
' per user reports

DMsg "generating user reports..."

For Each offender in offenders.Keys

	' reset file count for this user
	fileCount = 0

	' reset temp file
	Set tempFile = FSO.OpenTextFile(tempFileName, ForWriting, True)

	' build email address
	userEmail = offender & "@" & userDomain

	' write out body text
	tempFile.Write bodyText

	' prefix before file list
	tempFile.WriteLine
	tempFile.WriteLine "The following files were reported to be open:"

	' write out open file list
	For Each openFile in offenders.Item(offender).Keys
		fileCount = fileCount + 1
		tempFile.WriteLine openFile
	Next

	' summary count footer
	tempFile.WriteLine
	tempFile.WriteLine fileCount & " file(s)"

	' email tech footer
	tempFile.WriteLine
	tempFile.WriteLine "Generated " & Now
	tempFile.WriteLine "Generated on computer " & hostname
	tempFile.WriteLine "For user " & offender
	tempFile.WriteLine "For server " & fileServer
	tempFile.WriteLine "Admin contact: " & adminEmail

	tempFile.Close

	If Debugging Then
		sendTo = adminEmail
	Else
		sendTo = userEmail
	End If

	BlatMail userSubj, sendTo

Next ' offender

' ----------------------------------------------------------------------
' clean up

FSO.DeleteFile tempFileName

' ======================================================================
' END OF FILE
' ======================================================================
